@@ -140,12 +140,19 @@ struct evdev_device {
 	const char *devname;
 	bool was_removed;
 	int fd;
+	enum evdev_device_seat_capability seat_caps;
+	enum evdev_device_tags tags;
+	bool is_mt;
+	bool is_suspended;
+	int dpi; /* HW resolution */
+	struct ratelimit syn_drop_limit; /* ratelimit for SYN_DROPPED logging */
+	struct ratelimit nonpointer_rel_limit; /* ratelimit for REL_* events from non-pointer devices */
+	uint32_t model_flags;
+	struct mtdev *mtdev;
+
 	struct {
 		const struct input_absinfo *absinfo_x, *absinfo_y;
-		int fake_resolution;
-
-		struct device_coords point;
-		int32_t seat_slot;
+		bool is_fake_resolution;
 
 		int apply_calibration;
 		struct matrix calibration;
@@ -155,17 +162,6 @@ struct evdev_device {
 		struct device_coords dimensions;
 	} abs;
 
-	struct {
-		int slot;
-		struct mt_slot *slots;
-		size_t slots_len;
-		bool want_hysteresis;
-		struct device_coords hysteresis_margin;
-	} mt;
-	struct mtdev *mtdev;
-
-	struct device_coords rel;
-
 	struct {
 		struct libinput_timer timer;
 		struct libinput_device_config_scroll_method config;
@@ -193,24 +189,14 @@ struct evdev_device {
 		bool natural_scrolling_enabled;
 
 		/* angle per REL_WHEEL click in degrees */
-		int wheel_click_angle;
+		struct wheel_angle wheel_click_angle;
 	} scroll;
 
-	enum evdev_event_type pending_event;
-	enum evdev_device_seat_capability seat_caps;
-	enum evdev_device_tags tags;
-
-	int is_mt;
-	int suspended;
-
 	struct {
 		struct libinput_device_config_accel config;
 		struct motion_filter *filter;
 	} pointer;
 
-	/* Bitmask of pressed keys used to ignore initial release events from
-	 * the kernel. */
-	unsigned long hw_key_mask[NLONGS(KEY_CNT)];
 	/* Key counter used for multiplexing button events internally in
 	 * libinput. */
 	uint8_t key_count[KEY_CNT];
@@ -237,12 +223,6 @@ struct evdev_device {
 		uint32_t button_mask;
 		uint64_t first_event_time;
 	} middlebutton;
-
-	int dpi; /* HW resolution */
-	struct ratelimit syn_drop_limit; /* ratelimit for SYN_DROPPED logging */
-	struct ratelimit nonpointer_rel_limit; /* ratelimit for REL_* events from non-pointer devices */
-
-	uint32_t model_flags;
 };
 
 #define EVDEV_UNHANDLED_DEVICE ((struct evdev_device *) 1)
@@ -286,10 +266,24 @@ struct evdev_dispatch_interface {
 	 * was sent */
 	void (*post_added)(struct evdev_device *device,
 			   struct evdev_dispatch *dispatch);
+
+	void (*toggle_touch)(struct evdev_dispatch *dispatch,
+			     struct evdev_device *device,
+			     bool enable);
 };
 
 struct evdev_dispatch {
 	struct evdev_dispatch_interface *interface;
+
+	struct {
+		struct libinput_device_config_send_events config;
+		enum libinput_config_send_events_mode current_mode;
+	} sendevents;
+};
+
+struct fallback_dispatch {
+	struct evdev_dispatch base;
+
 	struct libinput_device_config_calibration calibration;
 
 	struct {
@@ -300,9 +294,29 @@ struct evdev_dispatch {
 	} rotation;
 
 	struct {
-		struct libinput_device_config_send_events config;
-		enum libinput_config_send_events_mode current_mode;
-	} sendevents;
+		struct device_coords point;
+		int32_t seat_slot;
+	} abs;
+
+	struct {
+		int slot;
+		struct mt_slot *slots;
+		size_t slots_len;
+		bool want_hysteresis;
+		struct device_coords hysteresis_margin;
+	} mt;
+
+	struct device_coords rel;
+
+	/* Bitmask of pressed keys used to ignore initial release events from
+	 * the kernel. */
+	unsigned long hw_key_mask[NLONGS(KEY_CNT)];
+
+	enum evdev_event_type pending_event;
+
+	/* true if we're reading events (i.e. not suspended) but we're
+	   ignoring them */
+	bool ignore_events;
 };
 
 struct evdev_device *
@@ -319,9 +333,9 @@ evdev_transform_relative(struct evdev_device *device,
 
 void
 evdev_init_calibration(struct evdev_device *device,
-		       struct evdev_dispatch *dispatch);
+		        struct libinput_device_config_calibration *calibration);
 
-int
+void
 evdev_device_init_pointer_acceleration(struct evdev_device *device,
 				       struct motion_filter *filter);
 
@@ -368,7 +382,7 @@ void
 evdev_device_calibrate(struct evdev_device *device,
 		       const float calibration[6]);
 
-int
+bool
 evdev_device_has_capability(struct evdev_device *device,
 			    enum libinput_device_capability capability);
 
@@ -417,7 +431,7 @@ double
 evdev_device_transform_y(struct evdev_device *device,
 			 double y,
 			 uint32_t height);
-int
+void
 evdev_device_suspend(struct evdev_device *device);
 
 int
@@ -429,12 +443,6 @@ evdev_notify_suspended_device(struct evdev_device *device);
 void
 evdev_notify_resumed_device(struct evdev_device *device);
 
-void
-evdev_keyboard_notify_key(struct evdev_device *device,
-			  uint64_t time,
-			  int key,
-			  enum libinput_key_state state);
-
 void
 evdev_pointer_notify_button(struct evdev_device *device,
 			    uint64_t time,
@@ -500,7 +508,7 @@ evdev_convert_to_mm(const struct input_absinfo *absinfo, double v)
 	return value/absinfo->resolution;
 }
 
-int
+void
 evdev_init_left_handed(struct evdev_device *device,
 		       void (*change_to_left_handed)(struct evdev_device *));
 
@@ -539,4 +547,59 @@ evdev_libinput_context(const struct evdev_device *device)
 	return device->base.seat->libinput;
 }
 
+/**
+ * Convert the pair of coordinates in device space to mm. This takes the
+ * axis min into account, i.e. a unit of min is equivalent to 0 mm.
+ */
+static inline struct phys_coords
+evdev_device_units_to_mm(const struct evdev_device* device,
+			 const struct device_coords *units)
+{
+	struct phys_coords mm = { 0,  0 };
+	const struct input_absinfo *absx, *absy;
+
+	if (device->abs.absinfo_x == NULL ||
+	    device->abs.absinfo_y == NULL) {
+		log_bug_libinput(evdev_libinput_context(device),
+				 "%s: is not an abs device\n",
+				 device->devname);
+		return mm;
+	}
+
+	absx = device->abs.absinfo_x;
+	absy = device->abs.absinfo_y;
+
+	mm.x = (units->x - absx->minimum)/absx->resolution;
+	mm.y = (units->y - absy->minimum)/absy->resolution;
+
+	return mm;
+}
+
+/**
+ * Convert the pair of coordinates in mm to device units. This takes the
+ * axis min into account, i.e. 0 mm  is equivalent to the min.
+ */
+static inline struct device_coords
+evdev_device_mm_to_units(const struct evdev_device *device,
+			 const struct phys_coords *mm)
+{
+	struct device_coords units = { 0,  0 };
+	const struct input_absinfo *absx, *absy;
+
+	if (device->abs.absinfo_x == NULL ||
+	    device->abs.absinfo_y == NULL) {
+		log_bug_libinput(evdev_libinput_context(device),
+				 "%s: is not an abs device\n",
+				 device->devname);
+		return units;
+	}
+
+	absx = device->abs.absinfo_x;
+	absy = device->abs.absinfo_y;
+
+	units.x = mm->x * absx->resolution + absx->minimum;
+	units.y = mm->y * absy->resolution + absy->minimum;
+
+	return units;
+}
 #endif /* EVDEV_H */
