@@ -71,6 +71,11 @@ struct normalized_range_coords {
 	double x, y;
 };
 
+/* A pair of angles in degrees */
+struct wheel_angle {
+	int x, y;
+};
+
 /* A pair of angles in degrees */
 struct tilt_degrees {
 	double x, y;
@@ -82,6 +87,12 @@ struct threshold {
 	int lower;
 };
 
+/* A pair of coordinates in mm */
+struct phys_coords {
+	double x;
+	double y;
+};
+
 struct tablet_axes {
 	struct device_coords point;
 	struct normalized_coords delta;
@@ -158,6 +169,11 @@ struct libinput_device_config_tap {
 	enum libinput_config_tap_state (*get_enabled)(struct libinput_device *device);
 	enum libinput_config_tap_state (*get_default)(struct libinput_device *device);
 
+	enum libinput_config_status (*set_map)(struct libinput_device *device,
+						   enum libinput_config_tap_button_map map);
+	enum libinput_config_tap_button_map (*get_map)(struct libinput_device *device);
+	enum libinput_config_tap_button_map (*get_default_map)(struct libinput_device *device);
+
 	enum libinput_config_status (*set_drag_enabled)(struct libinput_device *device,
 							enum libinput_config_drag_state);
 	enum libinput_config_drag_state (*get_drag_enabled)(struct libinput_device *device);
@@ -665,7 +681,7 @@ normalized_length(struct normalized_coords norm)
 	return hypot(norm.x, norm.y);
 }
 
-static inline int
+static inline bool
 normalized_is_zero(struct normalized_coords norm)
 {
 	return norm.x == 0.0 && norm.y == 0.0;
